#include <random>
#include <cmath>
#include <chrono>
#include <vector>
#include <iostream>
#include "FastGA.hpp"

int main(int argc, char const *argv[])
{
    auto ico = FastGA::Ico::IcoCharts(2);
    
    return 0;
}


