/* The hello_world namespace.
 *
 * Hello World implements a standard namespace with a few available functions.
 */
#ifndef FASTGA_HPP
#define FASTGA_HPP
#define _USE_MATH_DEFINES
#include <cmath> 

#include "FastGA/FastGA.hpp"
#include "FastGA/Helper.hpp"
#include "NanoS2ID/NanoS2ID.hpp"
#endif
